"""
Functions to perform workspace related operations
"""
import json
import os

from chenpy import client


def extract_dir(res):
    """Extract the directory name from the AtomQL response"""
    if res.get("response"):
        dir_name = res.get("response").split('String = "')[-1].split('"')[0]
        return dir_name
    return None


async def ls(connection):
    """Retrieves the list of Atoms in the workspace"""
    res = await client.q(connection, "workspace")
    if "[io.appthreat.chencli.console.ChenProject] = empty" in res.get("response", ""):
        return None
    return res.get("response")


async def slice_cpg(
    connection,
    src,
    out_dir=None,
    languages="autodetect",
    project_name=None,
    slice_mode="Usages",
):
    """Function to slice the Atom based on a slice mode"""
    app_manifests = await create_cpg(
        connection,
        src,
        out_dir=out_dir,
        languages=languages,
        project_name=project_name,
        slice=True,
        slice_mode=slice_mode,
    )
    if not app_manifests:
        return None
    if isinstance(app_manifests, (list, tuple)):
        app_manifests = app_manifests[0]
    if app_manifests.get("slice_out") and os.path.exists(
        app_manifests.get("slice_out")
    ):
        try:
            with open(app_manifests.get("slice_out"), encoding="utf-8") as fp:
                return json.load(fp)
        except json.JSONDecodeError:
            return None
    return None


async def create_atom(
    connection,
    src,
    out_dir=None,
    languages="autodetect",
    project_name=None,
):
    """Function to create atom using atomgen server"""
    return await create_cpg(
        connection,
        src,
        out_dir=out_dir,
        languages=languages,
        project_name=project_name,
        slice=False,
        slice_mode="Usages",
        auto_build=True,
        skip_sbom=False,
        use_atom=True,
    )


async def create_cpg(
    connection,
    src,
    out_dir=None,
    languages="autodetect",
    project_name=None,
    slice=None,
    slice_mode="Usages",
    auto_build=True,
    skip_sbom=True,
    use_atom=False,
):
    """Function to create Atom using atomgen server"""
    app_manifests = []
    res = await client.create_cpg(
        connection,
        src,
        out_dir=out_dir,
        lang=languages,
        slice=slice,
        slice_mode=slice_mode,
    )
    if res:
        app_manifests = res.get("app_manifests", [])
        if app_manifests:
            first_app = app_manifests[0]
            if not project_name and first_app.get("app"):
                project_name = first_app.get("app")
            cpg_path = first_app.get("cpg")
            res = await import_atom(connection, cpg_path, project_name)
            if not res:
                return None
    return (
        app_manifests[0] if app_manifests and len(app_manifests) == 1 else app_manifests
    )


async def import_atom(connection, cpg_path, project_name=None):
    """Function to import Atom"""
    if cpg_path:
        res = await dir_exists(connection, cpg_path)
        if not res:
            raise ValueError(
                f"Atom {cpg_path} doesn't exist for import into chen. Check if the directory containing this Atom is mounted and accessible from the server."
            )
    if project_name:
        res = await client.q(
            connection, f"""importCpg("{cpg_path}", "{project_name}")"""
        )
    else:
        res = await client.q(connection, f"""importCpg("{cpg_path}")""")
    if isinstance(res, str):
        return False
    return True


async def import_code(connection, directory, project_name=None, language=""):
    """Function to import code to chen"""
    if project_name:
        res = await client.q(
            connection,
            f"""importCode("{directory}", projectName="{project_name}", language="{language.upper()}")""",
        )
    else:
        res = await client.q(
            connection, f"""importCode("{directory}", language="{language.upper()}")"""
        )
    if isinstance(res, str):
        return False
    if "io.shiftleft.codepropertygraph.Cpg" in res.get("response", ""):
        return True
    return False


async def from_string(connection, code_snippet, language="jssrc"):
    """Function to import string"""
    res = await client.q(
        connection, f'importCode.{language}.fromString("""{code_snippet}""")'
    )
    if isinstance(res, str):
        return False
    if res and "Code successfully imported" in res.get("response", ""):
        return True
    return False


async def reset(connection):
    """Function to reset workspace"""
    await client.q(connection, "workspace.reset")
    return True


async def get_path(connection):
    """Function to retrieve the path to a workspace"""
    res = await client.q(connection, "workspace.getPath")
    return extract_dir(res)


async def get_active_project(connection):
    """Function to retrieve active project"""
    return await client.q(connection, "workspace.getActiveProject")


async def set_active_project(connection, project_name):
    """Function to set active project"""
    res = await client.q(
        connection, f"""workspace.setActiveProject("{project_name}")"""
    )
    if isinstance(res, str):
        return False
    if res and "projectFile = ProjectFile(inputPath =" in res.get("response", ""):
        return True
    return False


async def delete_project(connection, project_name):
    """Function to delete a project"""
    return await client.q(connection, f"""workspace.deleteProject("{project_name}")""")


async def cpg_exists(connection, project_name):
    """Function to check if a Atom exists in the workspace"""
    if project_name:
        res = await client.q(connection, f"""workspace.cpgExists("{project_name}")""")
        if res and "Boolean = true" in res.get("response", ""):
            return True
    return False


async def get_overlay_dir(connection, project_name):
    """Function to retrieve the overlays of a project"""
    res = await client.q(
        connection, f"""workspace.overlayDirByProjectName("{project_name}")"""
    )
    return extract_dir(res)


async def dir_exists(connection, dir_name):
    """Function to check if a directory exists and is accessible from the chen server"""
    res = await client.q(connection, f"""os.exists(os.Path("{dir_name}"))""")
    if res and "Boolean = true" in res.get("response", ""):
        return True
    return False



"""Common utilities"""
import base64
import mimetypes
import os
import re
import shutil
import sys
import tarfile
import tempfile
import zipfile
from hashlib import blake2b

import httpx
import orjson
import pkg_resources
import psutil
import rich.progress
from packageurl import PackageURL
from packageurl.contrib import purl2url
from psutil._common import bytes2human
from rich.console import Console
from rich.json import JSON
from rich.panel import Panel
from rich.progress import Progress
from rich.syntax import Syntax
from rich.table import Table
from rich.tree import Tree

from chenpy.source import ghsa

GIT_AVAILABLE = False
try:
    import git

    GIT_AVAILABLE = True
except ImportError:
    pass

MAVEN_CENTRAL_URL = "https://repo1.maven.org/maven2/"
ANDROID_MAVEN = "https://maven.google.com/"

mimetypes.init()

svmem = psutil.virtual_memory()
max_memory = bytes2human(getattr(svmem, "available"), format="%(value).0f%(symbol)s")
CPU_COUNT = str(psutil.cpu_count())

if not os.getenv("JAVA_OPTS"):
    os.environ["JAVA_OPTS"] = f"-Xmx{max_memory}"

only_bat_ext = ".bat" if sys.platform == "win32" else ""
bin_ext = ".bat" if sys.platform == "win32" else ".sh"
exe_ext = ".exe" if sys.platform == "win32" else ""
USE_SHELL = True if sys.platform == "win32" else False

console = Console(
    log_time=False,
    log_path=False,
    color_system="auto",
    width=int(os.getenv("COLUMNS", "180")),
)

check_labels_list = (
    "check",
    "valid",
    "sanit",
    "escape",
    "clean",
    "safe",
    "serialize",
    "convert",
    "authenticate",
    "authorize",
    "encode",
    "encrypt",
)

CHEN_DATAFLOW_TRACKED_WIDTH = os.getenv("CHEN_DATAFLOW_TRACKED_WIDTH", "128")
if (
    isinstance(CHEN_DATAFLOW_TRACKED_WIDTH, str)
    and CHEN_DATAFLOW_TRACKED_WIDTH.isdigit()
):
    CHEN_DATAFLOW_TRACKED_WIDTH = int(CHEN_DATAFLOW_TRACKED_WIDTH)


def t(result, title="", caption="", language="javascript"):
    """Function to print the result as a table"""
    return print_table(result, title, caption, language)


def print_table(result, title="", caption="", language="javascript"):
    if isinstance(result, Table):
        console.print(result)
        return
    """Function to print the result as a table"""
    table = Table(
        title=title,
        caption=caption,
        show_lines=True,
        expand=True,
        header_style="bold magenta",
    )
    cols_added = {}
    if isinstance(result, dict) and result.get("response"):
        console.print(result.get("response"))
    elif isinstance(result, list) and len(result):
        for row in result:
            if isinstance(row, dict):
                rows = []
                iterrows = []
                if row.get("_1"):
                    iterrows.append(row.get("_1"))
                if row.get("_2"):
                    iterrows.append(row.get("_2"))
                if not row.get("_1") and not row.get("_2"):
                    iterrows.append(row)
                for rowToUse in iterrows:
                    if isinstance(rowToUse, list):
                        if len(rowToUse):
                            if rowToUse[0].get("_label"):
                                if not cols_added.get(rowToUse[0].get("_label")):
                                    table.add_column(
                                        rowToUse[0].get("_label").lower(),
                                        overflow="fold",
                                    )
                                    cols_added[rowToUse[0].get("_label")] = True
                            elif rowToUse[0].get("name"):
                                if not cols_added.get(rowToUse[0].get("name")):
                                    table.add_column(
                                        rowToUse[0].get("name").lower(), overflow="fold"
                                    )
                                    cols_added[rowToUse[0].get("name")] = True
                            rows.append(JSON.from_data(rowToUse, default=""))
                    elif isinstance(rowToUse, dict):
                        for k, v in rowToUse.items():
                            # Simplify the table a bit for display
                            if k.startswith("ast") or k.startswith("column"):
                                continue
                            if not cols_added.get(k):
                                justify = "left"
                                if (
                                    k in ("id", "order")
                                    or "number" in k.lower()
                                    or "index" in k.lower()
                                ):
                                    justify = "right"
                                table.add_column(k, justify=justify, overflow="fold")
                                cols_added[k] = True
                            if isinstance(v, list) and len(v) == 0:
                                v = ""
                            if k == "code":
                                v = v.replace("<empty>", "")
                                rows.append(Syntax(v, language) if v else v)
                            elif k == "annotation" or isinstance(v, dict):
                                rows.append(JSON.from_data(v, default=""))
                            else:
                                rows.append(
                                    str(v).replace("<empty>", "")
                                    if v is not None
                                    else ""
                                )
                table.add_row(*rows)
            elif isinstance(row, list):
                table.add_row(row)
        console.print(table)
    else:
        console.print(result)


def print_md(result):
    """Function to print the result as a markdown"""
    if isinstance(result, dict) and result.get("response"):
        console.print(result.get("response"))
    else:
        console.print(result)


def walk_tree(paths, tree, level_branches):
    """Utility function to walk call tree"""
    added_nodes = []
    for path in paths:
        if not path:
            continue
        level = path.count("|    ")
        if level == 0:
            branch = tree
        elif level_branches.get(level - 1):
            branch = level_branches.get(level - 1)
            if not level_branches.get(level):
                level_branches[level] = branch
        n = path.replace("+--- ", "")
        if n not in added_nodes:
            branch.add(n)
            added_nodes.append(n)


def print_tree(result, guide_style="bold bright_blue"):
    """Function to print call trees"""
    result = result.split("\n")
    if result:
        tree = Tree(result[0].replace("+--- ", ""), guide_style=guide_style)
        level_branches = {0: tree}
        if len(result) > 1:
            walk_tree(result[1:], tree, level_branches)
        console.print(tree)
    else:
        console.print("Empty tree")


def calculate_hash(content, digest_size=16):
    """Function to calculate has using blake2b algorithm"""
    h = blake2b(digest_size=digest_size)
    if content and isinstance(content, str):
        content = content.replace("\n", "").replace("\t", "").replace(" ", "")
    h.update(content.encode())
    return h.hexdigest()


def print_flows(
    result,
    symbol_highlight_color="bold red",
    filelocation_highlight_color="grey54",
    check_highlight_color="dim green",
):
    """Function to print the data flows using a rich tree"""
    if not result:
        return
    parsed_flows_list = []
    for res in result:
        useful_flows = []
        identifiers_list = []
        ftree = None
        floc_list = []
        flow_fingerprint_list = []
        symbol = ""
        last_symbol = ""
        node_name = ""
        code = ""
        last_code = ""
        if isinstance(res, dict) and res.get("_2"):
            location_list = res.get("_2")
            for idx, loc in enumerate(location_list):
                # Allow empty sink
                filename = loc.get("filename", "")
                if filename == "<empty>" and idx < len(location_list) - 2:
                    continue
                symbol = loc.get("symbol", "").encode().decode("unicode_escape")
                node_name = loc.get("node", {}).get("name")
                if symbol.startswith("<operator"):
                    continue
                # Add the various computed fingerprints
                loc["fingerprints"] = {}
                code = (
                    (
                        loc.get("node", {})
                        .get("code", "")
                        .encode()
                        .decode("unicode_escape")
                        .strip()
                    )
                    .replace("    ", " ")
                    .replace("\t", "")
                    .replace("\n", " ")
                    .replace("    ", "")
                )
                if len(code) > CHEN_DATAFLOW_TRACKED_WIDTH:
                    code = code[:CHEN_DATAFLOW_TRACKED_WIDTH] + " ..."
                if code == last_code:
                    continue
                method_full_name = loc.get("methodFullName", "").replace("<init>", "")
                method_short_name = loc.get("methodShortName", "").replace("<init>", "")
                class_name = loc.get("className")
                # If there is no class name but there is a method full name try to recover
                if not class_name and method_full_name:
                    class_name = method_full_name.split(":")[0]
                    if class_name.endswith("." + method_short_name):
                        class_name = re.sub(f".{method_short_name}$", "", class_name)
                # Highlight potential check methods
                if check_highlight_color:
                    for check_label in check_labels_list:
                        if check_label in method_short_name:
                            method_short_name = method_short_name.replace(
                                method_short_name,
                                f"[{check_highlight_color}]{method_short_name}[/{check_highlight_color}]",
                            )
                        if node_name and node_name in code:
                            if check_label in node_name.lower():
                                code = code.replace(
                                    node_name,
                                    f"[{check_highlight_color}]{node_name}[/{check_highlight_color}]",
                                )
                if code == "<empty>":
                    code = ""
                for lk in ("methodShortName", "methodFullName", "symbol", "filename"):
                    if loc.get(lk):
                        loc["fingerprints"][lk] = calculate_hash(loc.get(lk))
                if code:
                    loc["fingerprints"]["code"] = calculate_hash(code)
                node_label = loc.get("nodeLabel")
                if node_label in ("METHOD_PARAMETER_IN", "CALL") or (
                    filename.endswith(".py")
                    and node_label == "IDENTIFIER"
                    and (idx == 0 or idx == len(location_list) - 1)
                ):
                    floc = f"{loc.get('filename')}:{loc.get('lineNumber')} {method_short_name}()"
                    floc_key = f"{floc}|{symbol}"
                    # If the next entry in the flow is identical to this
                    # but better then ignore the current
                    if idx < len(location_list) - 1:
                        nextloc = location_list[idx + 1]
                        next_node_label = nextloc.get("nodeLabel")
                        if next_node_label in (
                            "METHOD_PARAMETER_IN",
                            "CALL",
                            "IDENTIFIER",
                        ):
                            nextfloc = (
                                f"{nextloc.get('filename')}:{nextloc.get('lineNumber')}"
                            )
                            nextfloc_key = f"""{nextfloc}|{nextloc.get("symbol")}"""
                            if floc == nextfloc and (
                                floc_key == nextfloc_key
                                or len(floc_key) < len(nextfloc_key)
                            ):
                                continue
                    if loc.get("filename") == "<empty>":
                        class_method_sep = "" if not method_short_name else "."
                        if symbol_highlight_color:
                            floc = f"{class_name}{class_method_sep}{method_short_name}( [{symbol_highlight_color}]{code}[/{symbol_highlight_color}] )"
                        else:
                            floc = f"{class_name}{class_method_sep}{method_short_name}( {code} )"
                    if floc_key not in floc_list:
                        if symbol == code and last_symbol and last_symbol != code:
                            symbol = last_symbol
                        if (
                            symbol
                            and symbol_highlight_color
                            and (symbol in code or symbol in identifiers_list)
                        ):
                            last_code = code
                            code = (
                                code.replace(
                                    f'"{symbol}',
                                    f'"[{symbol_highlight_color}]{symbol}[/{symbol_highlight_color}]',
                                )
                                .replace(
                                    f"({symbol},",
                                    f"( [{symbol_highlight_color}]{symbol} [/{symbol_highlight_color}],",
                                )
                                .replace(
                                    f"{symbol})",
                                    f"[{symbol_highlight_color}]{symbol}[/{symbol_highlight_color}] )",
                                )
                                .replace(
                                    f"{symbol},",
                                    f"[{symbol_highlight_color}]{symbol}[/{symbol_highlight_color}] ,",
                                )
                                .replace(
                                    f"{symbol}[",
                                    f"[{symbol_highlight_color}]{symbol}[/{symbol_highlight_color}][",
                                )
                                .replace(
                                    f"&{symbol}",
                                    f"[{symbol_highlight_color}]&{symbol}[/{symbol_highlight_color}]",
                                )
                                .replace(
                                    f"*{symbol}",
                                    f"[{symbol_highlight_color}]*{symbol}[/{symbol_highlight_color}]",
                                )
                                .replace(
                                    f" {symbol} ",
                                    f" [{symbol_highlight_color}]{symbol}[/{symbol_highlight_color}] ",
                                )
                            )
                            last_symbol = symbol
                        if loc.get("filename") == "<empty>":
                            tree_content = floc
                        elif filelocation_highlight_color:
                            tree_content = f"[{filelocation_highlight_color}]{floc}[/{filelocation_highlight_color}] {code}"
                        else:
                            tree_content = f"{floc} {code}"
                        if not ftree:
                            ftree = Tree(tree_content)
                        else:
                            ftree.add(tree_content)
                        useful_flows.append(loc)
                        floc_list.append(floc_key)
                        if loc["fingerprints"].get("methodFullName"):
                            flow_fingerprint_list.append(
                                f'{loc["fingerprints"]["methodFullName"]}|{loc["fingerprints"]["symbol"]}'
                            )
                if (
                    node_label in ("METHOD_PARAMETER_IN", "IDENTIFIER")
                    and symbol not in identifiers_list
                    and not symbol.startswith("$")
                    and not symbol.startswith("tmp")
                    and not symbol.startswith("_tmp_")
                    and symbol != "NULL"
                ):
                    identifiers_list.append(symbol)
                    last_symbol = symbol
        if ftree:
            flow_fingerprint_key = "-".join(flow_fingerprint_list)
            if flow_fingerprint_key not in parsed_flows_list:
                if identifiers_list:
                    console.print(
                        Panel(
                            "\n".join(identifiers_list),
                            expand=False,
                            title="Tainted Identifiers",
                        )
                    )
                console.print(Panel(ftree, expand=False, title="Data Flow"))
                console.print("")
                parsed_flows_list.append(flow_fingerprint_key)


def expand_search_str(search_descriptor):
    """Given a descriptor string or dict, this method converts into equivalent cpgql method"""
    search_str = ""
    if isinstance(search_descriptor, str):
        if "." in search_descriptor:
            if (
                ":" in search_descriptor or "(" in search_descriptor
            ) and ".*" not in search_descriptor:
                search_str = f'.fullNameExact("{search_descriptor}")'
            else:
                search_str = f'.fullName("{search_descriptor}")'
        else:
            search_str = f'.name("{search_descriptor}")'
    elif isinstance(search_descriptor, dict):
        for k, v in search_descriptor.items():
            search_str = f'{search_str}.{k}("{v}")'
    return search_str


def fix_json(sout):
    """Hacky method to convert the chen stdout string to json"""
    source_sink_mode = False
    original_sout = sout
    try:
        if "defined function source" in sout:
            source_sink_mode = True
            sout = sout.replace("defined function source\n", "")
            sout = sout.replace("defined function sink\n", "")
        else:
            sout = sout.replace(r'"\"', '"').replace(r'\""', '"')
        if ': String = "[' in sout or ": String = [" in sout or source_sink_mode:
            if ": String = [" in sout:
                sout = sout.split(": String = ")[-1]
            elif source_sink_mode:
                sout = (
                    sout.replace(r"\"", '"')
                    .replace('}]}]"', "}]}]")
                    .replace('\\"', '"')
                )
                if ': String = "[' in sout:
                    sout = sout.split(': String = "')[-1]
            else:
                sout = sout.split(': String = "')[-1][-1]
        elif (
            "tree: ListBuffer" in sout
            or " = ListBuffer(" in sout
            or ': String = """' in sout
        ):
            sout = sout.split(": String = ")[-1]
            if '"""' in sout:
                sout = sout.replace('"""', "")
            return sout
        elif "me.shadaj.scalapy.py.Dynamic | Unit = " in sout:
            sout = sout.split("me.shadaj.scalapy.py.Dynamic | Unit = ")[-1]
            if '"""' in sout:
                sout = sout.replace('"""', "")
            return sout
        elif 'String = """[' in sout:
            tmpA = sout.split("\n")[1:-2]
            sout = "[ " + "\n".join(tmpA) + "]"
        sout = sout.replace('"}]"', '"}]')
        return orjson.loads(sout)
    except orjson.JSONDecodeError:
        return {"response": original_sout}


def fix_query(query_str):
    """Utility method to convert AtomQL queries to become json friendly"""
    if "\\." in query_str and "\\\\." not in query_str:
        query_str = query_str.replace("\\.", "\\\\.")
    if (query_str.startswith("atom.") or query_str.startswith("({atom.")) and (
        ".toJson" not in query_str
        and ".plotDot" not in query_str
        and not query_str.endswith(".p")
        and ".store" not in query_str
        and "def" not in query_str
        and "printCallTree" not in query_str
    ):
        query_str = f"{query_str}.toJsonPretty"
    if "\n" not in query_str and (
        os.getenv("POLYNOTE_VERSION") or os.getenv("AT_DEBUG_MODE") in ("true", "1")
    ):
        console.print(
            Panel(Syntax(query_str, "scala"), expand=False, title="AtomQL Query")
        )
    return query_str


def parse_error(serr):
    """Function to parse chen output and identify friendly error messages"""
    if "No projects loaded" in serr:
        return """ERROR: Import code using import_code api. Usage: await workspace.import_code(connection, directory_name, app_name)"""
    if "No Atom loaded" in serr:
        return """ERROR: Import atom using import_atom or import_code api."""
    return serr


def read_image(file_path):
    """Function to read image file safely optionally converting binary formats to base64 string. Useful to render
    images in notebooks"""
    if os.path.exists(file_path):
        (mt, encoding) = mimetypes.guess_type(file_path, strict=True)
        if mt.startswith("image/svg"):
            with open(file_path, mode="r", encoding=encoding) as fp:
                return fp.read()
        if mt.startswith("image/"):
            with open(file_path, mode="rb") as fp:
                b = base64.b64encode(fp.read())
                return b.decode("utf-8")
    return None


def colorize_dot_data(
    dot_data,
    scheme="paired9",
    colors=None,
    shapes=None,
    style="filled",
):
    """
    Function to colorize dot data with Brewer color schemes

    This product includes color specifications and designs developed by Cynthia Brewer (http://colorbrewer.org/).
    """
    if shapes is None:
        shapes = {
            "method": "box3d",
            "literal": "oval",
            "operator": "box",
            "param": "tab",
            "identifier": "note",
            "modifier": "rect",
            "type_ref": "component",
            "return": "cds",
        }
    if colors is None:
        colors = {
            "method": "1",
            "literal": "2",
            "operator": "3",
            "param": "4",
            "identifier": "5",
            "modifier": "6",
            "unknown": "7",
            "local": "7",
            "type_ref": "8",
            "return": "9",
        }
    fdot_list = []
    if dot_data and isinstance(dot_data, list):
        for d in dot_data:
            if "digraph" in d:
                for k, v in colors.items():
                    if k == "operator":
                        d = d.replace(
                            "label = <(&lt;operator&gt;",
                            f"color = {v}, shape = {shapes.get('operator', 'box')}, style = {style}, label = <(&lt;operator&gt;",
                        )
                    else:
                        d = d.replace(
                            f"label = <({k.upper()},",
                            f"color = {v}, shape = {shapes.get(k, 'ellipse')}, style = {style}, label = <({k.upper()},",
                        )
                d = d.replace("label = <", "label=<")
                afdot = d.split("\n")
                afdot.insert(1, f"node [colorscheme={scheme}];")
                fdot_list.append("\n".join(afdot))
            else:
                fdot_list.append(d)
        return fdot_list[0] if fdot_list and len(fdot_list) == 1 else fdot_list
    return dot_data


def get_version():
    """
    Returns the version of depscan
    """
    return pkg_resources.get_distribution("appthreat-chen").version


def unzip_unsafe(zf, to_dir):
    """Method to unzip the file in an unsafe manne"""
    with zipfile.ZipFile(zf, "r") as zip_ref:
        zip_ref.extractall(to_dir)
    try:
        os.remove(zf)
    except Exception:
        pass


def check_command(cmd):
    """
    Method to check if command is available
    :return True if command is available in PATH. False otherwise
    """
    cpath = shutil.which(cmd, mode=os.F_OK | os.X_OK)
    return cpath is not None


def is_binary_string(content):
    """
    Method to check if the given content is a binary string
    """
    textchars = bytearray({7, 8, 9, 10, 12, 13, 27} | set(range(0x20, 0x100)) - {0x7F})
    return bool(content.translate(None, textchars))


def is_exe(src):
    """Detect if the source is a binary file
    :param src: Source path
    :return True if binary file. False otherwise.
    """
    if os.path.isfile(src):
        try:
            return is_binary_string(open(src, "rb").read(1024))
        except OSError:
            return False
    return False


def import_url(src):
    """Method to import code from url"""
    if not os.path.exists(src):
        clone_dir = tempfile.mkdtemp(prefix="chen")
        if src.startswith("http") or src.startswith("git://"):
            clone_repo(src, clone_dir)
        else:
            download_package_unsafe(src, clone_dir)
        src = clone_dir
    return src


def clone_repo(repo_url, clone_dir, depth=1):
    """Method to clone a git repo"""
    if not GIT_AVAILABLE:
        return None
    git.Repo.clone_from(repo_url, clone_dir, depth=depth)
    return clone_dir


def build_maven_download_url(purl):
    """
    Return a maven download URL from the `purl` string.
    """
    url_prefix = MAVEN_CENTRAL_URL

    purl_data = PackageURL.from_string(purl)
    group = purl_data.namespace.replace(".", "/")
    name = purl_data.name
    version = purl_data.version

    if "android" in group:
        url_prefix = ANDROID_MAVEN

    if name and version:
        return f"{url_prefix}{group}/{name}/{version}/{name}-{version}.jar"


def build_pypi_download_url(purl):
    """
    Return a PyPI download URL from the `purl` string.
    """
    url_prefix = "https://pypi.io/packages/source/"
    purl_data = PackageURL.from_string(purl)
    name = purl_data.name
    version = purl_data.version
    if name and version:
        return f"{url_prefix}{name[0]}/{name}/{name}-{version}.tar.gz"


def build_golang_download_url(purl):
    """
    Return a golang download URL from the `purl` string.
    """
    purl_data = PackageURL.from_string(purl)
    namespace = purl_data.namespace
    name = purl_data.name
    version = purl_data.version
    qualifiers = purl_data.qualifiers
    download_url = qualifiers.get("download_url")
    if download_url:
        return download_url
    if not (namespace and name and version):
        return
    version_prefix = qualifiers.get("version_prefix", "v")
    version = f"{version_prefix}{version}"
    return f"https://{namespace}/{name}/archive/refs/tags/{version}.zip"


def build_ghsa_download_url(cve_or_ghsa):
    """Method to get download urls for the packages belonging to the CVE"""
    return ghsa.get_download_urls(cve_or_ghsa=cve_or_ghsa)


def get_download_url(purl_str):
    """Build download urls from a purl or CVE or GHSA id"""
    if purl_str.startswith("GHSA") or purl_str.startswith("CVE"):
        return build_ghsa_download_url(purl_str)
    if purl_str.startswith("pkg:maven"):
        return build_maven_download_url(purl_str)
    if purl_str.startswith("pkg:pypi"):
        return build_pypi_download_url(purl_str)
    if purl_str.startswith("pkg:golang"):
        return build_golang_download_url(purl_str)
    return purl2url.get_download_url(purl_str)


def untar_unsafe(tf, to_dir):
    """Method to untar .tar or .tar.gz files in an unsafe manner"""
    if tf.endswith("tar.gz") or tf.endswith(".tgz"):
        tar = tarfile.open(tf, "r:gz")
        tar.extractall(to_dir)
        tar.close()
    elif tf.endswith(".tar"):
        tar = tarfile.open(tf, "r:")
        tar.extractall(to_dir)
        tar.close()
    shutil.rmtree(tf, ignore_errors=True)


def download_package_unsafe(purl_str, download_dir, expand_archive=True):
    """Method to download the package from the given purl or CVE id"""
    if not purl_str:
        return
    durl = get_download_url(purl_str)
    if not durl:
        return
    if isinstance(durl, str):
        durl = [durl]
    for aurl in durl:
        if isinstance(aurl, dict) and aurl.get("purl"):
            aurl = get_download_url(aurl.get("purl"))
        with open(
            os.path.join(download_dir, os.path.basename(aurl)), mode="wb"
        ) as download_file:
            with httpx.stream("GET", aurl, follow_redirects=True) as response:
                total = int(response.headers["Content-Length"])
                with Progress(
                    "[progress.percentage]{task.percentage:>3.0f}%",
                    rich.progress.BarColumn(bar_width=None),
                    rich.progress.DownloadColumn(),
                    rich.progress.TransferSpeedColumn(),
                ) as progress:
                    download_task = progress.add_task("Download", total=total)
                    for chunk in response.iter_bytes():
                        download_file.write(chunk)
                        progress.update(
                            download_task, completed=response.num_bytes_downloaded
                        )
            download_file.close()
            if expand_archive:
                if download_file.name.endswith(".zip"):
                    unzip_unsafe(download_file.name, download_dir)
                elif (
                    download_file.name.endswith(".tar")
                    or download_file.name.endswith(".tar.gz")
                    or download_file.name.endswith(".tgz")
                ):
                    untar_unsafe(download_file.name, download_dir)
    return download_dir


def purl_to_friendly_name(purl_str):
    """Convert package url to a friendly name"""
    purl_data = PackageURL.from_string(purl_str)
    name = purl_data.name
    version = purl_data.version
    return f"{name}-{version}"



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import argparse
import os
import shutil
import subprocess
import sys

import oras.client
from rich.progress import Progress

import chenpy.config as config
from chenpy.client import ChenDistributionRegistry
from chenpy.logger import LOG, console
from chenpy.utils import USE_SHELL, check_command, get_version, max_memory, unzip_unsafe

try:
    os.environ["PYTHONIOENCODING"] = "utf-8"
except Exception:
    pass


def build_args():
    """
    Constructs command line arguments for the chennai tool
    """
    parser = argparse.ArgumentParser(
        description="Code Hierarchy Exploration Net (chen)."
    )
    parser.add_argument(
        "--download",
        action="store_true",
        default=False,
        dest="download",
        help="Download the latest chen distribution in platform specific "
        "user_data_dir",
    )
    parser.add_argument(
        "--with-science",
        action="store_true",
        default=False,
        dest="science_pack",
        help="Download the science pack",
    )
    parser.add_argument(
        "--server",
        action="store_true",
        default=False,
        dest="server_mode",
        help="Start chen server",
    )
    parser.add_argument(
        "-v",
        "--version",
        help="Display the version",
        action="version",
        version="%(prog)s " + get_version(),
    )
    return parser.parse_args()


def find_jars(lib_dir):
    jars = []
    for dirname, subdirs, files in os.walk(lib_dir):
        for filename in files:
            if filename.endswith(".jar"):
                jars.append(os.path.join(dirname, filename))
    return ":".join(jars)


def fix_envs():
    if not os.getenv("CHEN_HOME") or config.chen_home not in os.getenv("PATH"):
        py_version = "python" + sys.version[:4]
        # On Windows, there is no dot
        if sys.platform == "win32":
            py_version = py_version.replace(".", "")
        platform_dir = os.path.join(config.chen_home, "platform")
        platform_bin_dir = os.path.join(config.chen_home, "platform", "bin")
        pylib_dir, lpy_version = detect_python_lib_path()
        if lpy_version and py_version != lpy_version:
            py_version = lpy_version
        if sys.platform == "win32":
            LOG.info(
                "To run chennai console, set the following user environment variables:"
            )
        else:
            LOG.info(
                "To run chennai console, add the following environment variables to your .zshrc or .bashrc:"
            )
        console.print(
            f'export JAVA_OPTS="-Xmx{max_memory}"\nexport JAVA_TOOL_OPTIONS="-Dfile.encoding=UTF-8 -Djna.library.path={pylib_dir}"\nexport SCALAPY_PYTHON_LIBRARY={py_version}\nexport CHEN_HOME="{config.chen_home}"\nexport PATH=$PATH{os.pathsep}"{platform_dir + os.pathsep + platform_bin_dir + os.pathsep}"'
        )
        if not os.getenv("JAVA_HOME"):
            LOG.info(
                "Ensure Java >= 21 is installed. Set the environment variable JAVA_HOME to point the correct "
                "java directory."
            )
        LOG.info(
            "After setting the values, restart the terminal and type chennai to launch the console."
        )
        os.environ[
            "JAVA_TOOL_OPTIONS"
        ] = f'"-Dfile.encoding=UTF-8 -Djna.library.path={pylib_dir}"'
        os.environ["SCALAPY_PYTHON_LIBRARY"] = py_version
        os.environ["CHEN_HOME"] = config.chen_home
        os.environ["CLASSPATH"] = (
            find_jars(os.path.join(config.chen_home, "platform", "lib"))
            + os.pathsep
            + os.getenv("CLASSPATH", "")
        )
        os.environ["PATH"] = (
            os.environ["PATH"] + platform_dir + os.pathsep + platform_bin_dir
        )


def download_chen_distribution(overwrite=False, science_pack=False):
    if os.path.exists(os.path.join(config.chen_home, "platform")):
        if not overwrite:
            if science_pack:
                install_py_modules("science")
            fix_envs()
            return
        LOG.debug(
            "Existing chen distribution at %s would be overwritten", config.chen_home
        )
    LOG.info(
        "About to download chen distribution from %s", config.chen_distribution_url
    )
    req_files = []
    oras_client = oras.client.OrasClient(registry=ChenDistributionRegistry())
    paths_list = oras_client.pull(
        target=config.chen_distribution_url,
        outdir=config.chen_home,
        allowed_media_type=[],
        overwrite=overwrite,
    )
    if paths_list:
        LOG.debug("Extracting chen to %s", config.chen_home)
        zip_files = [file for file in paths_list if file.endswith(".zip")]
        req_files = [file for file in paths_list if file.endswith(".txt")]
        if not zip_files:
            return
        unzip_unsafe(zip_files[0], config.chen_home)
        # Add execute permissions
        for dirname, subdirs, files in os.walk(config.chen_home):
            for filename in files:
                if (
                    not filename.endswith(".zip")
                    and not filename.endswith(".jar")
                    and not filename.endswith(".json")
                    and not filename.endswith(".dll")
                    and (
                        filename.endswith(".sh")
                        or "2cpg" in filename
                        or "chen" in filename
                        or "repl" in filename
                    )
                ):
                    try:
                        os.chmod(os.path.join(dirname, filename), 0o755)
                    except Exception:
                        pass
    # Install the science pack
    if req_files:
        install_py_modules("database")
    if science_pack:
        install_py_modules("science")
    fix_envs()


def install_py_modules(pack="database"):
    """
    Install the required science modules
    """
    if check_command("conda"):
        LOG.info(
            "About to install the science pack using conda. A new environment called 'chenpy-local' will be created."
        )
        with Progress(transient=True) as progress:
            conda_install_script = """conda create --name chenpy-local python=3.12 -y
conda install -n chenpy-local -c conda-forge networkx -y
conda install -n chenpy-local -c pytorch pytorch torchtext cpuonly -y
conda install -n chenpy-local -c conda-forge numpy packageurl-python nbconvert jupyter_core jupyter_client notebook -y
conda install -n chenpy-local -c conda-forge oras-py==0.1.26 httpx websockets orjson rich appdirs psutil gitpython -y"""
            for line in conda_install_script.split("\n"):
                if line.strip():
                    task = progress.add_task(line, start=False, total=100)
                    subprocess.check_call(
                        line.split(" "),
                        stdout=subprocess.DEVNULL,
                        stderr=subprocess.DEVNULL,
                        shell=USE_SHELL,
                    )
                    progress.stop_task(task)
                    progress.update(task, completed=100)
        LOG.info(
            "Activate the conda environment using the command 'conda activate chenpy-local'"
        )
    else:
        LOG.info("About to install the science pack using cpu-only configuration")
        req_file = os.path.join(config.chen_home, f"chen-{pack}-requirements.txt")
        if os.path.exists(req_file):
            subprocess.check_call(
                [sys.executable, "-m", "pip", "install", "-r", req_file],
                stdout=subprocess.DEVNULL,
                stderr=subprocess.DEVNULL,
                shell=USE_SHELL,
            )


def detect_python_lib_path():
    lib_dir = ""
    py_version = ""
    if sys.platform == "win32":
        python_exe_path = shutil.which("python")
        if python_exe_path:
            lib_dir = os.sep.join(python_exe_path.split(os.sep)[:-1])
            if lib_dir.endswith("Scripts") or os.getenv("VIRTUAL_ENV"):
                LOG.info(
                    "Unable to identify the directory containing python3.dll. Set the value for jna.library.path "
                    "in JAVA_TOOL_OPTIONS environment variable correctly."
                )
                lib_dir = ""
        return lib_dir, py_version
    cp = subprocess.run(
        ["python3-config", "--ldflags", "--embed"],
        check=True,
        capture_output=True,
        encoding="utf-8",
        shell=USE_SHELL,
    )
    if cp.returncode != 0:
        LOG.info("Unable to identify the python lib directory using python3-config")
        return lib_dir, py_version
    stdout = cp.stdout
    if stdout:
        tmpA = stdout.split(" ")
        if tmpA and len(tmpA) > 2:
            if tmpA[0].startswith("-L"):
                lib_dir = os.sep.join(tmpA[0].replace("-L", "").split(os.sep)[:-2])
            if tmpA[1].startswith("-l"):
                py_version = tmpA[1].replace("-l", "")
    return lib_dir, py_version


def main():
    """
    Detects the project type, performs various scans and audits,
    and generates reports based on the results.
    """
    args = build_args()
    download_chen_distribution(args.download, args.science_pack)


if __name__ == "__main__":
    main()



"""
Module to work with networkx graph data
"""
import json
import tempfile
from collections import Counter, defaultdict

from chenpy.utils import calculate_hash

DATABASE_PACK_AVAILABLE = True
SCIENCE_PACK_AVAILABLE = True
try:
    import networkx as nx
    from networkx.readwrite import json_graph, read_graphml
except ImportError:
    DATABASE_PACK_AVAILABLE = False

try:
    import torch
    from torch import Tensor
    from torch_geometric.data import Data
    from torchtext.data.functional import (
        generate_sp_model,
        load_sp_model,
        sentencepiece_numericalizer,
        sentencepiece_tokenizer,
    )
except ImportError:
    SCIENCE_PACK_AVAILABLE = False


def _hash_label(label, digest_size):
    return calculate_hash(label, digest_size=digest_size)


def _init_node_labels(G, edge_attr_fn, node_attr_fn):
    if node_attr_fn:
        return {u: node_attr_fn(dd) for u, dd in G.nodes(data=True)}
    elif edge_attr_fn:
        return {u: "" for u in G}
    else:
        return {u: str(deg) for u, deg in G.degree()}


def _neighborhood_aggregate(G, node, node_labels, edge_attr_fn=None):
    """
    Compute new labels for given node by aggregating
    the labels of each node's neighbors.
    """
    label_list = []
    for nbr in G.neighbors(node):
        prefix = "" if edge_attr_fn is None else edge_attr_fn(G[node][nbr])
        label_list.append(prefix + node_labels[nbr])
    return node_labels[node] + "".join(sorted(label_list))


def graph_hash(G, edge_attr_fn=None, node_attr_fn=None, iterations=3, digest_size=16):
    """Return Weisfeiler Lehman (WL) graph hash"""

    def weisfeiler_lehman_step(G, labels, edge_attr_fn=None):
        """
        Apply neighborhood aggregation to each node
        in the graph.
        Computes a dictionary with labels for each node.
        """
        new_labels = {}
        for node in G.nodes():
            label = _neighborhood_aggregate(G, node, labels, edge_attr_fn=edge_attr_fn)
            new_labels[node] = _hash_label(label, digest_size)
        return new_labels

    # set initial node labels
    node_labels = _init_node_labels(G, edge_attr_fn, node_attr_fn)

    subgraph_hash_counts = []
    for _ in range(iterations):
        node_labels = weisfeiler_lehman_step(G, node_labels, edge_attr_fn=edge_attr_fn)
        counter = Counter(node_labels.values())
        # sort the counter, extend total counts
        subgraph_hash_counts.extend(sorted(counter.items(), key=lambda x: x[0]))

    # hash the final counter
    return _hash_label(str(tuple(subgraph_hash_counts)), digest_size)


def subgraph_hashes(
    G, edge_attr_fn=None, node_attr_fn=None, iterations=3, digest_size=16
):
    """Return a dictionary of subgraph hashes by node."""

    def weisfeiler_lehman_step(G, labels, node_subgraph_hashes, edge_attr_fn=None):
        """
        Apply neighborhood aggregation to each node
        in the graph.
        Computes a dictionary with labels for each node.
        Appends the new hashed label to the dictionary of subgraph hashes
        originating from and indexed by each node in G
        """
        new_labels = {}
        for node in G.nodes():
            label = _neighborhood_aggregate(G, node, labels, edge_attr_fn=edge_attr_fn)
            hashed_label = _hash_label(label, digest_size)
            new_labels[node] = hashed_label
            node_subgraph_hashes[node].append(hashed_label)
        return new_labels

    node_labels = _init_node_labels(G, edge_attr_fn, node_attr_fn)

    node_subgraph_hashes = defaultdict(list)
    for _ in range(iterations):
        node_labels = weisfeiler_lehman_step(
            G, node_labels, node_subgraph_hashes, edge_attr_fn
        )

    return dict(node_subgraph_hashes)


def diff(first_graph, second_graph, include_common=False, as_dict=False, as_dot=False):
    """Function to compute the difference between two graphs"""
    return diff_graph(
        first_graph,
        second_graph,
        include_common=include_common,
        as_dict=as_dict,
        as_dot=as_dot,
    )


def get_node_label(n):
    """Retrieve a label for the node from various data attributes"""
    if not n:
        return ""
    for at in (
        "label",
        "CODE",
        "SIGNATURE",
        "METHOD_FULL_NAME",
        "NAME",
        "VARIABLE",
        "labelE",
    ):
        if n.get(at) is not None:
            return n.get(at)
    return ""


def diff_graph(
    first_graph, second_graph, include_common=False, as_dict=False, as_dot=False
):
    """Function to compute the difference between two graphs and optionally convert the result to dict or dot format"""
    if not first_graph and second_graph:
        return second_graph
    if first_graph and not second_graph:
        return first_graph
    graph = nx.Graph()
    if not first_graph and not second_graph:
        return graph
    first_graph_nodes = [get_node_label(r[1]) for r in first_graph.nodes(data=True)]
    second_graph_nodes = [get_node_label(r[1]) for r in second_graph.nodes(data=True)]
    removed_nodes = set(first_graph_nodes) - set(second_graph_nodes)
    added_nodes = set(second_graph_nodes) - set(first_graph_nodes)
    nodes = set(second_graph_nodes) & set(first_graph_nodes)
    first_graph_edges = [
        (r[0], r[1], get_node_label(r[2])) for r in first_graph.edges(data=True)
    ]
    second_graph_edges = [
        (r[0], r[1], get_node_label(r[2])) for r in second_graph.edges(data=True)
    ]
    removed_edges = set(first_graph_edges) - set(second_graph_edges)
    removed_edges_fmt = []
    for removed_edge in removed_edges:
        src = removed_edge[0]
        dest = removed_edge[1]
        if removed_edge[0] in removed_nodes:
            src = "-" + removed_edge[0]
        if removed_edge[1] in removed_nodes:
            dest = "-" + removed_edge[1]
        removed_edges_fmt.append((src, dest, removed_edge[2]))
    added_edges = set(second_graph_edges) - set(first_graph_edges)
    added_edges_fmt = []
    for added_edge in added_edges:
        src = added_edge[0]
        dest = added_edge[1]
        if added_edge[0] in added_nodes:
            src = "+" + added_edge[0]
        if added_edge[1] in added_nodes:
            dest = "+" + added_edge[1]
        added_edges_fmt.append((src, dest, added_edge[2]))
    edges = set(second_graph_edges) & set(first_graph_edges)
    for removed_node in removed_nodes:
        graph.add_node("-" + removed_node)
    for added_node in added_nodes:
        graph.add_node("+" + added_node)
    if include_common:
        for node in nodes:
            graph.add_node(node)
    for removed_edge in removed_edges_fmt:
        graph.add_edge(removed_edge[0], removed_edge[1], label="-" + removed_edge[2])
    for added_edge in added_edges_fmt:
        graph.add_edge(added_edge[0], added_edge[1], label="+" + added_edge[2])
    if include_common:
        for edge in edges:
            graph.add_edge(edge[0], edge[1], label=edge[2])
    if as_dict:
        return nx.to_dict_of_dicts(graph)
    if as_dot:
        with tempfile.NamedTemporaryFile(
            prefix="diff_graph", suffix=".dot", delete=False
        ) as fp:
            write_dot(graph, fp.name)
            fp.flush()
            return fp.read().decode()
    return graph


def node_match_fn(n1, n2):
    return get_node_label(n1) == get_node_label(n2)


def gep(first_graph, second_graph, upper_bound=500):
    distance = nx.optimal_edit_paths(
        first_graph,
        second_graph,
        node_match=node_match_fn,
        edge_match=node_match_fn,
        upper_bound=upper_bound,
    )
    if distance is None:
        distance = -1
    return distance


def ged(first_graph, second_graph, timeout=5, upper_bound=500):
    distance = nx.graph_edit_distance(
        first_graph,
        second_graph,
        node_match=node_match_fn,
        edge_match=node_match_fn,
        timeout=timeout,
        upper_bound=upper_bound,
    )
    if distance is None:
        distance = -1
    return distance


def write_dot(G, path):
    """Function to export graph as dot"""
    nx.nx_agraph.write_dot(G, path)


def hash(
    G,
    subgraph=False,
    edge_attr_fn=get_node_label,
    node_attr_fn=get_node_label,
    iterations=3,
    digest_size=16,
):
    """Function to compute the hashes for a graph using Weisfeiler Lehman hashing algorithm"""
    if subgraph:
        return subgraph_hashes(
            G,
            edge_attr_fn=edge_attr_fn,
            node_attr_fn=node_attr_fn,
            iterations=iterations,
            digest_size=digest_size,
        )
    return graph_hash(
        G,
        edge_attr_fn=edge_attr_fn,
        node_attr_fn=node_attr_fn,
        iterations=iterations,
        digest_size=digest_size,
    )


def summarize(G, as_dict=False, as_dot=False):
    """Function to summarize the graph based on node labels"""
    summary_graph = nx.snap_aggregation(
        G, node_attributes=("label", "CODE"), edge_attributes=("label", "CODE")
    )
    if as_dict:
        return nx.to_dict_of_dicts(summary_graph)
    if as_dot:
        with tempfile.NamedTemporaryFile(
            prefix="summary_graph", suffix=".dot", delete=False
        ) as fp:
            write_dot(summary_graph, fp.name)
            fp.flush()
            return fp.read().decode()
    return summary_graph


def is_similar(M1, M2, edit_distance=10, upper_bound=500, timeout=5):
    if not diff_graph(M1, M2, as_dict=True):
        return True
    distance = ged(M1, M2, upper_bound=upper_bound, timeout=timeout)
    if distance == -1:
        return False
    return int(distance) < edit_distance


def convert_graphml(
    gml_file, force_multigraph=False, as_graph=True, as_adjacency_data=False
):
    """Function to convert graphml to networkx"""
    if not DATABASE_PACK_AVAILABLE:
        return RuntimeError(
            "Graph database dependencies missing. Please refer to the documentation to install the database pack or use the official chen container image."
        )
    try:
        G = read_graphml(gml_file, force_multigraph=force_multigraph)
        if as_graph:
            return G
        if as_adjacency_data:
            return json.dumps(json_graph.adjacency_data(G))
    except Exception:
        return None


def to_pyg(
    G,
    group_node_attrs=None,
    group_edge_attrs=None,
):
    """Converts a :obj:`networkx.Graph` or :obj:`networkx.DiGraph` to a
    :class:`torch_geometric.data.Data` instance."""
    if not SCIENCE_PACK_AVAILABLE:
        return RuntimeError(
            "Scientific dependencies missing. Please refer to the documentation to install the science pack or use the official chen container image."
        )
    G = G.to_directed() if not nx.is_directed(G) else G

    mapping = dict(zip(G.nodes(), range(G.number_of_nodes())))
    edge_index = torch.empty((2, G.number_of_edges()), dtype=torch.long)
    for i, (src, dst) in enumerate(G.edges()):
        edge_index[0, i] = mapping[src]
        edge_index[1, i] = mapping[dst]

    data = defaultdict(list)

    if G.number_of_nodes() > 0:
        node_attrs = list(next(iter(G.nodes(data=True)))[-1].keys())
    else:
        node_attrs = {}

    if G.number_of_edges() > 0:
        edge_attrs = list(next(iter(G.edges(data=True)))[-1].keys())
    else:
        edge_attrs = {}

    for i, (_, feat_dict) in enumerate(G.nodes(data=True)):
        for key, value in feat_dict.items():
            data[str(key)].append(value)

    for i, (_, _, feat_dict) in enumerate(G.edges(data=True)):
        for key, value in feat_dict.items():
            key = f"edge_{key}" if key in node_attrs else key
            data[str(key)].append(value)

    for key, value in G.graph.items():
        if key == "node_default" or key == "edge_default":
            continue  # Do not load default attributes.
        key = f"graph_{key}" if key in node_attrs else key
        data[str(key)] = value

    for key, value in data.items():
        if isinstance(value, (tuple, list)) and isinstance(value[0], Tensor):
            data[key] = torch.stack(value, dim=0)
        else:
            try:
                data[key] = torch.tensor(value)
            except (ValueError, TypeError):
                pass

    data["edge_index"] = edge_index.view(2, -1)
    data = Data.from_dict(data)

    if group_node_attrs is all:
        group_node_attrs = list(node_attrs)
    if group_node_attrs is not None:
        xs = []
        for key in group_node_attrs:
            x = data.get(key, "")
            x = x.view(-1, 1) if x.dim() <= 1 else x
            xs.append(x)
            if data.get(key) is not None:
                del data[key]
        data.x = torch.cat(xs, dim=-1)

    if group_edge_attrs is all:
        group_edge_attrs = list(edge_attrs)
    if group_edge_attrs is not None:
        xs = []
        for key in group_edge_attrs:
            key = f"edge_{key}" if key in node_attrs else key
            x = data.get(key, "")
            x = x.view(-1, 1) if x.dim() <= 1 else x
            xs.append(x)
            if data.get(key) is not None:
                del data[key]
        data.edge_attr = torch.cat(xs, dim=-1)

    if data.x is None and data.pos is None:
        data.num_nodes = G.number_of_nodes()

    return data



"""Client for chen server"""
import asyncio
import json
import os
import platform
import tempfile

import httpx
import oras.client
import oras.provider
from oras.logger import setup_logger

setup_logger(quiet=True, debug=False)

UVLOOP_AVAILABLE = True
try:
    import uvloop
except Exception:
    UVLOOP_AVAILABLE = False
import websockets

from chenpy.utils import (
    check_labels_list,
    expand_search_str,
    fix_json,
    fix_query,
    parse_error,
    print_flows,
    print_md,
    print_table,
)

if platform.system() == "Windows":
    asyncio.set_event_loop_policy(asyncio.WindowsSelectorEventLoopPolicy())
elif UVLOOP_AVAILABLE:
    asyncio.set_event_loop_policy(uvloop.EventLoopPolicy())


headers = {"Content-Type": "application/json", "Accept-Encoding": "gzip"}
CLIENT_TIMEOUT = os.getenv("HTTP_CLIENT_TIMEOUT")


class Connection:
    """
    Connection object to hold following connections:
       - Websocket to chen server
       - http connection to chen server
       - http connection to atomgen server
    """

    def __init__(self, atomgenclient, httpclient, websocket):
        self.atomgenclient = atomgenclient
        self.httpclient = httpclient
        self.websocket = websocket

    async def __aenter__(self):
        return self

    async def ping(self):
        """Send websocket ping message"""
        await self.websocket.ping()

    async def close(self):
        """Close all connections"""
        await self.atomgenclient.close()
        await self.httpclient.close()
        await self.websocket.close()

    async def __aexit__(self, exc_type, exc_value, exc_traceback):
        return await self.close()


def get_sync(
    base_url="http://localhost:9000",
    atomgen_url="http://localhost:7072",
    username=None,
    password=None,
):
    """Function to create a plain synchronous http connection to chen and atomgen server"""
    auth = None
    if username and password:
        auth = httpx.BasicAuth(username, password)
    base_url = base_url.rstrip("/")
    client = httpx.Client(base_url=base_url, auth=auth, timeout=CLIENT_TIMEOUT)
    atomgenclient = None
    if atomgen_url:
        atomgenclient = httpx.Client(base_url=atomgen_url, timeout=CLIENT_TIMEOUT)
    return Connection(atomgenclient, client, None)


async def get(
    base_url="http://localhost:8080",
    atomgen_url="http://localhost:7072",
    username=None,
    password=None,
):
    """Function to create a connection to chen and atomgen server"""
    auth = None
    if username and password:
        auth = httpx.BasicAuth(username, password)
    base_url = base_url.rstrip("/")
    client = httpx.AsyncClient(base_url=base_url, auth=auth, timeout=CLIENT_TIMEOUT)
    atomgenclient = None
    if atomgen_url:
        atomgenclient = httpx.AsyncClient(base_url=atomgen_url, timeout=CLIENT_TIMEOUT)
    ws_url = f"""{base_url.replace("http://", "ws://").replace("https://", "wss://")}/connect"""
    websocket = await websockets.connect(ws_url, ping_interval=None, ping_timeout=None)
    connected_msg = await websocket.recv()
    if connected_msg != "connected":
        raise websockets.exceptions.InvalidState(
            "Didn't receive connected message from chen server"
        )
    # Workaround to fix websockets.exceptions.ConnectionClosedError
    await asyncio.sleep(0)
    return Connection(atomgenclient, client, websocket)


async def send(connection, message):
    """Send message to the chen server via websocket"""
    await connection.websocket.send(message)


async def receive(connection):
    """Receive message from the chen server"""
    return await connection.websocket.recv()


async def p(connection, query_str, title="", caption="", sync=False):
    """Function to print the result as a table"""
    result = await query(connection, query_str, sync=sync)
    print_table(result, title, caption)
    return result


async def q(connection, query_str, sync=False):
    """Query chen server and optionally print the result as a table if the query ends with .p"""
    if query_str.strip().endswith(".p"):
        query_str = f"{query_str[:-2]}.toJsonPretty"
        return await p(connection, query_str, sync=sync)
    return await query(connection, query_str, sync=sync)


async def query(connection, query_str, sync=False):
    """Query chen server"""
    client = connection.httpclient
    if isinstance(client, httpx.AsyncClient):
        response = await client.post(
            url=f"/query{'-sync' if sync else ''}",
            headers=headers,
            json={"query": fix_query(query_str)},
        )
    else:
        sync = True
        response = client.post(
            url=f"/query{'-sync' if sync else ''}",
            headers=headers,
            json={"query": fix_query(query_str)},
        )
    if response.status_code == httpx.codes.OK:
        j = response.json()
        if sync:
            return fix_json(j.get("stdout", ""))
        res_uuid = j.get("uuid")
        try:
            completed_uuid = await receive(connection)
            if completed_uuid == res_uuid:
                response = await client.get(
                    url=f"/result/{completed_uuid}", headers=headers
                )
                if response.status_code == httpx.codes.OK:
                    j = response.json()
                    sout = j.get("stdout", "")
                    serr = j.get("stderr", "")
                    if sout:
                        return fix_json(sout)
                    return parse_error(serr)
        except Exception:
            return None
    return None


async def bulk_query(connection, query_list, sync=False):
    """Bulk query chen server"""
    client = connection.httpclient
    websocket = connection.websocket
    uuid_list = []
    response_list = []
    for query_str in query_list:
        if isinstance(client, httpx.AsyncClient):
            response = await client.post(
                url=f"/query{'-sync' if sync else ''}",
                headers=headers,
                json={"query": fix_query(query_str)},
            )
        else:
            sync = True
            response = client.post(
                url="/query-sync", headers=headers, json={"query": fix_query(query_str)}
            )
        if response.status_code == httpx.codes.OK:
            j = response.json()
            if sync:
                sout = j.get("stdout", "")
                serr = j.get("stderr", "")
                if sout:
                    response_list.append(fix_json(sout))
                else:
                    response_list.append({"error": parse_error(serr)})
            else:
                res_uuid = j.get("uuid")
                uuid_list.append(res_uuid)
    if sync:
        return response_list
    async for completed_uuid in websocket:
        if completed_uuid in uuid_list:
            response = await client.get(
                url=f"/result/{completed_uuid}", headers=headers
            )
            if response.status_code == httpx.codes.OK:
                j = response.json()
                sout = j.get("stdout", "")
                serr = j.get("stderr", "")
                if sout:
                    response_list.append(fix_json(sout))
                else:
                    response_list.append({"error": parse_error(serr)})
        if len(response_list) == len(uuid_list):
            return response_list
    return response_list


async def flows(connection, source, sink):
    """Execute reachableByFlows query"""
    return await flowsp(
        connection,
        source,
        sink,
        print_result=bool(os.getenv("POLYNOTE_VERSION")),
    )


async def flowsp(connection, source, sink, print_result=True):
    """Execute reachableByFlows query and optionally print the result table"""
    if not source.startswith("def"):
        source = f"def source = {source}"
    if not sink.startswith("def"):
        sink = f"def sink = {sink}"
    results = await bulk_query(
        connection,
        [
            source,
            sink,
            "sink.reachableByFlows(source).p",
        ],
    )
    if print_result and len(results):
        tmpres = results[-1]
        if isinstance(tmpres, dict) and tmpres.get("response"):
            tmpres = tmpres.get("response")
        tmpA = tmpres.split('"""')[1:-1]
        print_md("\n".join([n for n in tmpA if len(n.strip()) > 1]))
    return results


async def df(
    connection,
    source,
    sink,
    print_result=True if os.getenv("POLYNOTE_VERSION") else False,
    filter=None,
    check_labels=check_labels_list,
):
    """
    Execute reachableByFlows query. Optionally accepts filters which could be a raw conditional string or predefined keywords such as skip_control_structures, skip_cfg and skip_checks
    skip_control_structures: This adds a control structure filter `filter(m => m.elements.isControlStructure.size > 0)` to skip flows with control statements such if condition or break
    skip_cfg: This adds a cfg filter `filter(m => m.elements.isCfgNode.size > 0)` to skip flows with control flow graph nodes
    skip_checks: When used with check_labels parameter, this could filter flows containing known validation and sanitization code in the flow. Has a default list.
    """
    filter_str = ""
    if isinstance(check_labels, str):
        check_labels = check_labels.split("|")
    if isinstance(source, dict):
        for k, v in source.items():
            if k in ("parameter", "tag"):
                source = f"""atom.tag.name("{v}").{k}"""
            elif k in ("method", "call", "annotation"):
                source = f"""atom.{k}{expand_search_str(v)}"""
    if isinstance(sink, dict):
        for k, v in sink.items():
            if k in ("parameter", "tag"):
                sink = f"""atom.tag.name("{v}").{k}"""
            elif k in ("method", "call", "annotation"):
                sink = f"""atom.{k}{expand_search_str(v)}"""
    if not source.startswith("def"):
        source = f"def source = {source}"
    if not sink.startswith("def"):
        sink = f"def sink = {sink}"
    if filter:
        if isinstance(filter, str):
            if filter == "skip_checks":
                filter_str = f""".filter(m => m.elements.code(".*({'|'.join(check_labels)}).*").size == 0)"""
            elif not filter.startswith("."):
                filter_str = f".filter({filter})"
        elif isinstance(filter, list):
            for k in filter:
                if k == "skip_control_structures":
                    filter_str = f"{filter_str}.filter(m => m.elements.isControlStructure.size > 0)"
                elif k == "skip_cfg":
                    filter_str = (
                        f"{filter_str}.filter(m => m.elements.isCfgNode.size > 0)"
                    )
                elif k == "skip_checks":
                    filter_str = f"""{filter_str}.filter(m => m.elements.code(".*({'|'.join(check_labels)}).*").size == 0)"""
                else:
                    filter_str = f"""{filter_str}.filter({k})"""
    with tempfile.NamedTemporaryFile(
        prefix="reachable_flows_", suffix=".json", delete=False
    ) as fp:
        res = await bulk_query(
            connection,
            [
                source,
                sink,
                f'sink.reachableByFlows(source){filter_str}.map(m => (m, m.elements.location.l)).toJson |> "{fp.name}"',
            ],
        )
        try:
            results = json.load(fp)
            if print_result:
                print_flows(results)
        except json.JSONDecodeError:
            return res[-1] if len(res) else res
    return results


async def reachable_by_flows(connection, source, sink, print_result=False):
    """Execute reachableByFlows query"""
    return await df(connection, source, sink, print_result)


async def create_cpg(
    connection,
    src,
    out_dir=None,
    lang=None,
    slice=None,
    slice_mode="Usages",
    auto_build=True,
    skip_sbom=True,
):
    """Create Atom using atomgen server"""
    client = connection.atomgenclient
    if not client:
        return {
            "error": "true",
            "message": "No active connection to atomgen server. Pass the atomgen url to the client.get method.",
        }, 500
    # Suppor for url
    url = ""
    if src.startswith("http") or src.startswith("git://") or src.startswith("pkg:"):
        url = src
        src = ""
    data = {
        "src": src,
        "url": url,
        "out_dir": out_dir,
        "lang": lang,
        "slice": "true" if slice else "false",
        "slice_mode": slice_mode,
        "auto_build": "true" if auto_build else "false",
        "skip_sbom": "true" if skip_sbom else "false",
    }
    if isinstance(client, httpx.AsyncClient):
        response = await client.post(
            url="/cpg",
            headers=headers,
            json=data,
        )
    else:
        response = client.post(
            url="/cpg",
            headers=headers,
            json=data,
        )
    return response.json()


async def graphml_export(connection, filter_str="method"):
    """Method to export method as graphml"""
    with tempfile.NamedTemporaryFile(
        prefix="graphml_export_", suffix=".graphml", delete=False
    ) as fp:
        res = await bulk_query(
            connection,
            [
                """
import overflowdb.formats.ExportResult
import overflowdb.formats.graphml.GraphMLExporter
import java.nio.file.{Path, Paths}

case class MethodSubGraph(methodName: String, nodes: Set[Node]) {
  def edges: Set[Edge] = {
    for {
      node <- nodes
      edge <- node.bothE.asScala
      if nodes.contains(edge.inNode) && nodes.contains(edge.outNode)
    } yield edge
  }
}

def plus(resultA: ExportResult, resultB: ExportResult): ExportResult = {
  ExportResult(
    nodeCount = resultA.nodeCount + resultB.nodeCount,
    edgeCount = resultA.edgeCount + resultB.edgeCount,
    files = resultA.files ++ resultB.files,
    additionalInfo = resultA.additionalInfo
  )
}

def splitByMethod(cpg: Cpg): IterableOnce[MethodSubGraph] = {
  atom.%(filter_str)s.map { method =>
    MethodSubGraph(methodName = method.name, nodes = method.ast.toSet)
  }
}
"""
                % dict(filter_str=filter_str),
                """
({splitByMethod(cpg).iterator
  .map { case subGraph @ MethodSubGraph(methodName, nodes) =>
    GraphMLExporter.runExport(nodes, subGraph.edges, Paths.get("%(gml_file)s"))
  }
  .reduce(plus)
})
"""
                % dict(gml_file=fp.name),
            ],
        )
        return fp.name


class ChenDistributionRegistry(oras.provider.Registry):
    def get_manifest(self, container, allowed_media_type=None):
        """
        Retrieve a manifest for a package.

        :param container:  parsed container URI
        :type container: oras.container.Container or str
        :param allowed_media_type: one or more allowed media types
        :type allowed_media_type: str
        """
        if not allowed_media_type:
            allowed_media_type = [oras.defaults.default_manifest_media_type]
        headers = {"Accept": ";".join(allowed_media_type)}

        get_manifest = f"{self.prefix}://{container.manifest_url()}"  # type: ignore
        response = self.do_request(get_manifest, "GET", headers=headers)
        self._check_200_response(response)
        manifest = response.json()
        return manifest



import os

import httpx

# GitHub advisory feed url
ghsa_api_url = os.getenv("GITHUB_GRAPHQL_URL", "https://api.github.com/graphql")
api_token = os.getenv("GITHUB_TOKEN")
headers = {"Authorization": f"token {api_token}"}

ecosystem_type_dict = {
    "go": "golang",
    "rust": "cargo",
    "pip": "pypi",
    "rubygems": "gem",
}


def get_query(cve_or_ghsa=None, only_malware=False, extra_clause=None):
    """Method to construct the graphql query"""
    extra_args = ""
    if not cve_or_ghsa:
        extra_args = "first: 100"
    else:
        id_type = "GHSA" if cve_or_ghsa.startswith("GHSA") else "CVE"
        extra_args = (
            'first: 100, identifier: {type: %(id_type)s, value: "%(cve_or_ghsa)s"}'
            % dict(id_type=id_type, cve_or_ghsa=cve_or_ghsa)
        )
    if only_malware:
        extra_args = f"{extra_args}, classifications:MALWARE"
    if extra_clause:
        extra_args = f"{extra_args}, {extra_clause}"
    gqljson = {
        "query": """
            {
                securityAdvisories(
                    %(extra_args)s
                ) {
                    nodes {
                    id
                    ghsaId
                    summary
                    description
                    identifiers {
                        type
                        value
                    }
                    origin
                    publishedAt
                    updatedAt
                    references {
                        url
                    }
                    severity
                    withdrawnAt
                    vulnerabilities(first: 10) {
                        nodes {
                        firstPatchedVersion {
                            identifier
                        }
                        package {
                            ecosystem
                            name
                        }
                        severity
                        updatedAt
                        vulnerableVersionRange
                        }
                    }
                    }
                }
            }
        """
        % dict(extra_args=extra_args)
    }
    return gqljson


def parse_response(json_data):
    """Parse json response and convert to list of purls"""
    purl_list = []
    for node in (
        json_data.get("data", {}).get("securityAdvisories", {}).get("nodes", {})
    ):
        ghsa_id = node.get("ghsaId")
        vulnerable_nodes = node.get("vulnerabilities", {}).get("nodes", [])
        for vn in vulnerable_nodes:
            pkg = vn.get("package", {})
            version = ""
            if vn.get("firstPatchedVersion"):
                version = vn.get("firstPatchedVersion", {}).get("identifier", "")
            elif vn.get("vulnerableVersionRange"):
                version = vn.get("vulnerableVersionRange").split(" ")[-1]
            if pkg:
                ptype = pkg.get("ecosystem", "").lower()
                pname = pkg.get("name", "").lower().replace(":", "/")
                # This is the fixed version
                if ptype and pname and version:
                    purl = (
                        f"pkg:{ecosystem_type_dict.get(ptype, ptype)}/{pname}@{version}"
                    )
                    purl_list.append(
                        {
                            "ghsaId": ghsa_id,
                            "purl": purl,
                        }
                    )
    return purl_list


def get_download_urls(cve_or_ghsa=None, only_malware=False):
    """Method to get download urls for the packages belonging to the CVE"""
    if not api_token:
        raise ValueError("GITHUB_TOKEN is required with read:packages scope")
    client = httpx.Client(http2=True, follow_redirects=True, timeout=180)
    r = client.post(
        url=ghsa_api_url,
        json=get_query(cve_or_ghsa=cve_or_ghsa, only_malware=only_malware),
        headers=headers,
    )
    json_data = r.json()
    return parse_response(json_data)



"""
Module to detect common functions used in python applications
"""
import re

from chenpy import client

ROUTE_DECORATORS = "route"


async def list_decorator_location(connection, decorator, is_method_ref=True):
    """
    Function to list the methods and their location for a given decorator
    """
    method_ref_filter = ".isMethodRef" if is_method_ref else ""
    return await client.q(
        connection,
        f'atom.call.code(".*{decorator}.*").inCall.assignment.argument{method_ref_filter}.location',
    )


async def list_dict_assignment_location(connection, key=None, is_literal=True):
    """
    Function to list locations where a dictionary key is assigned a hardcoded value
    """
    code_filter = f'.code(".*{key}.*")' if key else ""
    literal_filter = ".isLiteral" if is_literal else ""
    return await client.q(
        connection,
        f"atom.call(Operators.indexAccess){code_filter}.inCall.assignment.argument{literal_filter}.location",
    )


def expand_decorators(rows):
    """Function to expand decorators to identify http methods and routes"""
    ret_rows = []
    for row in rows:
        m = {}
        ann = {}
        if not row or not isinstance(row, dict):
            continue
        if row.get("_1"):
            decorator_data = row.get("_1")
            for k, v in decorator_data.items():
                ann[k] = v
                if k == "code":
                    route_matches = re.search(r'"(/?.[^,\s()])+"', v)
                    if route_matches.group():
                        ann["routePattern"] = route_matches.group().replace('"', "")
                        for hm in ("GET", "DELETE", "PUT", "POST", "PATCH", "OPTION"):
                            if hm in v:
                                ann["httpMethod"] = hm
            m["annotation"] = ann
        if row.get("_2"):
            method_data = row.get("_2")
            for k, v in method_data.items():
                m[k] = v
        ret_rows.append(m)
    return ret_rows


async def list_http_routes(connection, decorators=ROUTE_DECORATORS):
    """Function to list http routes in the application"""
    res = await client.q(
        connection,
        f"""atom.call.code(".*{decorators}.*").inCall.assignment.map(m => (m, m.method))""",
    )
    return expand_decorators(res)



"""
Module to detect common functions used in java applications
"""
import re

from chenpy import client

HTTP_ANNOTATIONS = "org\\.springframework\\.web\\.bind\\.annotation\\..*"

FILTER_ANNOTATIONS = "javax\\.servlet\\.annotation\\.WebFilter"

SOURCE_TYPE_PATTERN = "(?i).*(cloud|framework|data|http|net|socket|io|security|text|xml|json|proto|rpc|java).*"
SOURCE_FILE_PATTERN = "(?i).*(controller|service).*"
SINK_TYPE_PATTERN = "(?i).*(cloud|framework|data|http|net|socket|io|security|text|xml|json|proto|rpc|java).*"


def expand_annotations(rows):
    """Function to expand annotation nodes by identifying http methods and routes used"""
    ret_rows = []
    for r in rows:
        m = {}
        if not r or not isinstance(r, dict):
            continue
        if r.get("_1"):
            method_data = r.get("_1")
            for k, v in method_data.items():
                m[k] = v
        if r.get("_2"):
            annotation_list = r.get("_2")
            mannotation_list = []
            for annotation_data in annotation_list:
                http_method = ""
                if annotation_data.get("_label") == "ANNOTATION":
                    if (
                        "org.springframework.web.bind.annotation"
                        in annotation_data.get("fullName")
                    ):
                        http_method = (
                            annotation_data.get("name").replace("Mapping", "").upper()
                        )
                        annotation_data["httpMethod"] = http_method
                    code = annotation_data.get("code")
                    route_matches = re.search(r'"(/?.)+"', code)
                    if route_matches:
                        annotation_data["routePattern"] = route_matches.group().replace(
                            '"', ""
                        )
                mannotation_list.append(annotation_data)
            if mannotation_list:
                m["annotation"] = (
                    mannotation_list[0]
                    if len(mannotation_list) == 1
                    else mannotation_list
                )
        ret_rows.append(m)
    return ret_rows


async def list_http_routes(connection, annotations=HTTP_ANNOTATIONS):
    """Retrieves the http routes in the application"""
    res = await client.q(
        connection,
        f"""atom.method.internal.where(_.annotation.fullName("{annotations}")).map(m => (m, m.annotation.l))""",
    )
    return expand_annotations(res)


async def list_http_filters(connection, annotations=FILTER_ANNOTATIONS):
    """Retrieves the http filters in the application"""
    res = await client.q(
        connection,
        f"""atom.typeDecl.where(_.annotation.fullName("{annotations}")).map(m => (m, m.annotation.l))""",
    )
    return expand_annotations(res)


async def list_unresolved_external_methods(connection):
    """Retrieves the methods without types resolved"""
    return await client.q(
        connection,
        """atom.method.external.where(_.fullName(".*<unresolved.*")).whereNot(_.name(".*<(operator|init)>.*"))""",
    )


async def list_methods(
    connection,
    modifier="public ",
    include_annotations=True,
    external=False,
    unresolved=True,
):
    """Function to retrieve list of methods"""
    external_bool_str = "external" if external else "internal"
    filter_str = '.whereNot(_.name(".*<(operator|init)>.*"))'
    if not unresolved:
        filter_str = f'{filter_str}.whereNot(_.fullName(".*<unresolved.*"))'
    if include_annotations:
        res = await client.q(
            connection,
            f"""atom.method.{external_bool_str}{filter_str}.code("{modifier}.*").map(m => (m, m.annotation.l))""",
        )
        return expand_annotations(res)
    else:
        return await client.q(
            connection,
            f"""atom.method.{external_bool_str}{filter_str}.code("{modifier}.*")""",
        )


def get_sources_query(
    pattern=SOURCE_TYPE_PATTERN,
    file_pattern=SOURCE_FILE_PATTERN,
    parameter_filter='typeFullName("java.lang.String")',
):
    """Construct a AtomQL query to list the sources for the application"""
    if parameter_filter and not parameter_filter.startswith("."):
        parameter_filter = f".{parameter_filter}"
    return f"""
        (atom.method.internal.where(_.annotation.fullName("{pattern}")).whereNot(_.fullName(".*<unresolved.*")).whereNot(_.name(".*<(operator|init)>.*")).parameter{parameter_filter} ++ atom.method.internal.where(_.filename("{file_pattern}")).whereNot(_.fullName(".*<unresolved.*")).whereNot(_.name(".*<(operator|init)>.*")).parameter{parameter_filter}).location.toJson
        """


async def list_sources(
    connection,
    pattern=SOURCE_TYPE_PATTERN,
    file_pattern=SOURCE_FILE_PATTERN,
    parameter_filter='typeFullName("java.lang.String")',
):
    """Retrieves the list of sources by filtering based on conventions"""
    if parameter_filter and not parameter_filter.startswith("."):
        parameter_filter = f".{parameter_filter}"
    return await client.q(
        connection,
        get_sources_query(
            pattern=pattern,
            file_pattern=file_pattern,
            parameter_filter=parameter_filter,
        ),
    )


def get_sinks_query(pattern=SINK_TYPE_PATTERN):
    """Construct a AtomQL query to list the sinks for the application"""
    return f'atom.method.external.where(_.fullName("{pattern}")).whereNot(_.fullName(".*<unresolved.*")).whereNot(_.name(".*<(operator|init)>.*")).whereNot(_.signature("boolean.*")).parameter.location'


async def list_sinks(connection, pattern=SINK_TYPE_PATTERN):
    """Retrieves the list of sinks by filtering based on conventions"""
    return await client.q(connection, get_sinks_query(pattern=pattern))


async def suggest_flows(
    connection,
):
    """Suggest some data flows by identifying common sources and sinks"""
    return await client.df(
        connection,
        get_sources_query().replace(".location.toJson", ""),
        get_sinks_query().replace(".location", ""),
    )
